/*
     Copyright (C) 2006-2008  Georges Da Costa

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

// #defineDEBUG2
// #defineDEBUG

#include <stdio.h>

#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#define FT_FREETYPE_H  <freetype/freetype.h>
#include <X11/Xlib.h>
#include <X11/Xft/Xft.h>

#include <X11/Xutil.h>

#include <X11/Xatom.h>

#include <pthread.h> 


static pthread_mutex_t mutex1=PTHREAD_MUTEX_INITIALIZER;
int lock=0;

typedef struct win_t {
  Window win;
  int old_x,old_y;
  unsigned int win_size_x;
  unsigned int win_size_y;
  int cx;
  int cy;

  Pixmap buff;

  unsigned int nb_fonts;
  XftFont **fonts;

  unsigned int nb_colors;
  XftColor *xft_colors;

  XftDraw *xft_draw;

  struct button* button_list;
  
  void (*draw)(struct win_t*, void*);
  void (*handler)(void*);
  void *data;

  struct win_t* next;
} win_t;

struct button {
  int x,y,x2,y2,type;
  void (*handler)(unsigned int, unsigned int, unsigned int, void*);
  void* argument;
  struct button* next;
};

#define GOSD_C
#include "gosd.h"
#include "toolbox.h"

static Display* display;		/* pointer to X Display structure.           */
static win_t *ww;

void gosd_print(win_t* w, const int x, const int y,
		const unsigned int font, const unsigned int color,
		const char* txt){
DEBUGIN2("gosd_draw_text %s",txt);
  XGlyphInfo info;
  XftTextExtents8(display, w->fonts[font], (XftChar8*) txt, strlen(txt), &info);

  XftDrawString8(w->xft_draw, &w->xft_colors[color],
		 w->fonts[font], x+info.x, y+info.y,
		 (XftChar8 *) txt, strlen(txt));
DEBUGOUT2("gosd_draw_text %s",txt);
}; 

unsigned int gosd_text_width(win_t* w, unsigned int font, const char*txt) {
  XGlyphInfo info;
  XftTextExtents8(display, w->fonts[font], (XftChar8*) txt, strlen(txt), &info);
  return info.width;
}

unsigned int gosd_text_height(win_t* w, unsigned int font, const char*txt){
  XGlyphInfo info;
  XftTextExtents8(display, w->fonts[font], (XftChar8*) txt, strlen(txt), &info);
  return info.height;
}


void gosd_add_font(win_t* w, const unsigned int index, const double size,
		   const char *name) {
  if(index+1 > w->nb_fonts) {
    w->nb_fonts=index+1;
    w->fonts = realloc(w->fonts, (w->nb_fonts)*sizeof(XftFont*));
  }
  w->fonts[index] = XftFontOpen(display, DefaultScreen (display),
			     XFT_FAMILY, XftTypeString, "sans",
			     XFT_SIZE, XftTypeDouble, size,
			     NULL);
}

void gosd_add_color(win_t* w, const unsigned int index,
		    const unsigned int r, const unsigned int g, const unsigned int b) {
  XRenderColor color;
  if(index+1 > w->nb_colors) {
    w->nb_colors = index+1;
    w->xft_colors = realloc(w->xft_colors,w->nb_colors*sizeof(XftColor));
  }

  color.alpha = 65535;
  color.red   = r;
  color.green = g;
  color.blue  = b;

  XftColorAllocValue (display,
		      DefaultVisual(display, DefaultScreen (display)),
		      DefaultColormap(display, DefaultScreen (display)),
		      &color,
		      &w->xft_colors[index]);
}



int gosd_lock_display() {
  pthread_mutex_lock( &mutex1 );
  if(lock==1) {
    pthread_mutex_unlock( &mutex1 );
    return -1;
  }
DEBUGIN("Lock display");
  lock=1;
  pthread_mutex_unlock( &mutex1 );
  return 0;
}

void gosd_release_display() {
  pthread_mutex_lock( &mutex1 );
DEBUGOUT("Lock display");
  lock=0;
  pthread_mutex_unlock( &mutex1 );
}

void handle_expose(const Display* display, win_t *w,
		   const XExposeEvent* expose_event) {

DEBUGIN("Handle_expose");
  if(gosd_lock_display(w) == -1) {
DEBUGOUT("Handle_expose locked");
    return;
  }

  gosd_update_display(w);
DEBUGOUT("Handle_expose");

  gosd_release_display(w);
}



void gosd_update_display(win_t *w) {
DEBUGIN2("update display %p", w)
  XCopyArea(display, w->buff, w->win,
	    DefaultGC(display, DefaultScreen(display)),
  	    0, 0, w->win_size_x, w->win_size_y, 0, 0);
  XFlush(display);
DEBUGOUT2("update display %p",w)
}

void gosd_put_background(win_t *w) {
    Atom p, type;
    int format;
    unsigned long items, after;
    unsigned char* data;

    p = XInternAtom( display, "ESETROOT_PMAP_ID", True );

    if(p == None) {
      GC gc=DefaultGC(display, DefaultScreen(display));
      XSetFillStyle(display, gc, FillSolid);
      XFillRectangle(display, w->buff, gc,
		     0,0,w->win_size_x,w->win_size_y);
    } else {

      XGetWindowProperty (display, DefaultRootWindow(display), p, 0L, 1L,
			  False, XA_PIXMAP, &type, &format,
			  &items, &after, &data);

      XCopyArea(display, *(Pixmap*)data, w->buff,
		DefaultGC(display, DefaultScreen(display)),
		w->old_x, w->old_y, w->win_size_x, w->win_size_y, 0, 0);

      free(data);

    }
}

void gosd_position(const int posx, const int posy, win_t *w) {
  w->old_x = posx;
  w->old_y = posy;
  XMoveWindow(display, w->win, posx, posy);  
}

void gosd_add_button(const unsigned int x, const unsigned int y,
		     const unsigned int w, const unsigned int h,
		     const unsigned int type,
		     void (func)(unsigned int, unsigned int, unsigned int, void*),
		     void* arg,
		     win_t *win) {

  struct button* next = win->button_list;

  win->button_list = malloc(sizeof(struct button));
  win->button_list->x=x;
  win->button_list->y=y;
  win->button_list->x2=x+w;
  win->button_list->y2=y+h;
  win->button_list->type=type;
  win->button_list->handler=func;
  win->button_list->argument=arg;

  win->button_list->next=next;
}

void release(const XButtonEvent* event, win_t *w) {
  w->cx=-1;
}

void motion(const XButtonEvent* event, win_t *w) {
  int x = event->x;
  int y = event->y;
  int dx,dy;

  if(w->cx == -1)
    return;

  dx = x-w->cx;
  dy = y-w->cy;

  gosd_position(w->old_x+dx, w->old_y+dy, w);  
}

void button(const XButtonEvent* event, win_t *w) {
  int x = event->x;
  int y = event->y;
  int b = event->button;

  struct button* current = w->button_list;

  while(current != NULL) {
    if(current->x<=x && current->y<=y &&
       current->x2>=x && current->y2>= y &&
       current->type == b) {

      (*(current->handler))(x-current->x,y-current->y, b, current->argument);

      /* Refresh the screen to update according to the new state */
      w->draw(w,w->data);
      return;
    }
    current = current->next;
  }
  // else : we move
  if(b == 1) {
    w->cx=x;
    w->cy=y;
  }
}

void gosd_add_timeout(void(*handler)(void*), void* data) {
  win_t *res;
  res=malloc(sizeof(win_t));
  res->next=ww;
  res->draw=NULL;
  res->handler=handler;
  res->data=data;
  ww=res;
}


void draw_all(const unsigned int alarm) {
  win_t *ptr;
  ptr=ww;
  int first = TRUE;
  while(ptr != NULL) {

    if(ptr->draw != NULL) {
      if(first) {
	XEvent ev;

	ev.type = ColormapNotify;
	ev.xcolormap.display = display;
	ev.xcolormap.window = ptr->win;
	ev.xcolormap.colormap = None;
	ev.xcolormap.send_event=True;
	ev.xcolormap.state = ColormapUninstalled;

	XSendEvent(display, ptr->win, TRUE, 0L, &ev);
	XFlush(display);
	
	first=FALSE;
      }

      //   ptr->draw(ptr, ptr->data);
    } else if(alarm && ptr->handler != NULL) {
      ptr->handler(ptr->data);
    }
    ptr=ptr->next;
  }
}

void draw_all_draw(const unsigned int alarm) {
  win_t *ptr;
  ptr=ww;
  while(ptr != NULL) {
    if(ptr->draw != NULL) {
      ptr->draw(ptr, ptr->data);
    } else if(alarm && ptr->handler != NULL) {
      ptr->handler(ptr->data);
    }
    ptr=ptr->next;
  }
}

void alarm_f(const int a) {
DEBUGIN("alarm")
  draw_all(TRUE);
DEBUGOUT("alarm")
  alarm(1);
}

void gosd_set_handler(void (*handler)(win_t*, void*), void* data,win_t *w) {
  w->draw=handler;
  w->data=data;
}

void gosd_loop() {
  XEvent an_event;
  win_t *current_win;

  alarm_f(0);
  signal(SIGALRM, alarm_f);
  alarm(1);


  while (TRUE) {
DEBUGIN("Wait Event");
    XNextEvent(display, &an_event);

    current_win = ww;
    while(current_win != NULL && current_win->win != an_event.xany.window) {
      current_win = current_win->next;      
    }
    if(current_win==NULL) {
DEBUGOUT("Wait Event no win");
      continue;
    }
DEBUGOUT("Wait Event");

DEBUGIN2("Treat XNextEvent %p",current_win);

    switch (an_event.type) {
    case Expose:
      handle_expose(display, current_win, (XExposeEvent*)&an_event.xexpose);
      break;
    case ConfigureNotify:
      current_win->old_x = ((XConfigureEvent*)&an_event)->x;
      current_win->old_y = ((XConfigureEvent*)&an_event)->y;
      break;
    case ColormapNotify:
DEBUGIN2("Colormap %p",current_win);
      draw_all_draw(FALSE);
DEBUGOUT2("Colormap %p",current_win);
      break;
    case MotionNotify:
      motion((XButtonEvent*)&an_event, current_win);
      break;
    case ButtonRelease:
      release((XButtonEvent*)&an_event,current_win);
      break;
    case ButtonPress:
DEBUGIN2("button %p", current_win);
      button((XButtonEvent*)&an_event,current_win);
DEBUGOUT2("button %p", current_win);
      break;
    default: /* ignore any other event types. */
      break;
    } /* end switch on event type */

DEBUGOUT2("Treat XNextEvent %p",current_win);
    
  } /* end while events handling */
  XCloseDisplay(display);
}

void gosd_size(const unsigned int sizex, const unsigned int sizey, win_t *w) {
  w->win_size_x = sizex;
  w->win_size_y = sizey;
  XResizeWindow(display, w->win, sizex, sizey);

  gosd_position(w->old_x, w->old_y, w);  

  XFreePixmap(display, w->buff);
  w->buff = XCreatePixmap(display, w->win, w->win_size_x, w->win_size_y,
		       ScreenOfDisplay(display, DefaultScreen (display))->root_depth);
  XftDrawChange(w->xft_draw, w->buff);
}

win_t *gosd_new_window(const int posx, const int posy,
		       const unsigned int sizex, const unsigned int sizey,
		       const char *name) {
  XSetWindowAttributes attr;
  win_t *res;
  res=malloc(sizeof(win_t));
  res->next=ww;
  ww=res;

  res->cx=-1;
  res->button_list=NULL;
  res->draw=NULL;
  res->old_x = posx; res->old_y=posy;
  res->win_size_x = sizex;
  res->win_size_y = sizey;
  res->nb_fonts=0;
  res->fonts=NULL;
  res->nb_colors=0;
  res->xft_colors=NULL;
  res->handler=NULL;
  res->data=NULL;

  attr.event_mask = StructureNotifyMask | Button1MotionMask | ExposureMask
                  | ButtonPressMask | ButtonReleaseMask | ColormapChangeMask;

  res->win = XCreateWindow(display, DefaultRootWindow(display),
		      posx, posy, sizex, sizey, 0,
		      CopyFromParent, InputOutput, CopyFromParent,
		      CWEventMask, &attr);
  //  if(type == ALWAYS_VISIBLE)  {
  /*
    Atom atom = XInternAtom(display,"ATOM",0);
    Atom wm   = XInternAtom(display, "_NET_WM_WINDOW_TYPE", True);
    Atom val = XInternAtom(display, "_NET_WM_WINDOW_TYPE_DESKTOP",0);
    XChangeProperty(display, res->win, wm, atom, 32,
		    PropModeReplace,
		    (unsigned char*) &val, 1);
  */
  // reserve some place : 4x32
    /*  {
    Atom atom = XInternAtom(display,"CARDINAL",0);
    Atom wm   = XInternAtom(display, "_NET_WM_STRUT", True);

    Atom val[4] = { 8, 8, 16, 16 };

    XChangeProperty(display, win, wm, atom, 32,
		    PropModeReplace,
		    &val, 1);
		    }*/
    //  }


  
  {
    XWMHints wmhints;
    wmhints.input = False;
    wmhints.flags = InputHint ;
    XSetWMHints(display, res->win, &wmhints);
  }
  

  {
    XSizeHints hints;
    //    hints.x = posx;
    //    hints.y = posy;
    hints.flags = USPosition;
    XSetStandardProperties (display, res->win, NULL, NULL, (Pixmap) 0,
			    NULL, 0, &hints);
  }

    /*
  {
    Atom atom = XInternAtom(display,"ATOM",0);
    Atom wm   = XInternAtom(display, "_NET_WM_ALLOWED_ACTIONS", True);
    Atom val = XInternAtom(display, "_NET_WM_ACTION_MOVE",0);
    XChangeProperty(display, win, wm, atom, 32,
		    PropModeReplace,
		    &val, 1);
  }
    */

 

  // no borders, and may be sticky too ?
{
    Atom WM_HINTS;
    WM_HINTS = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    {
      unsigned char test[] = { 0x02, 0x00, 0x00, 0x00,
			       0x00, 0x00, 0x00, 0x00,
			       0x00, 0x00, 0x00, 0x00,
			       0xff, 0xff, 0xff, 0xff,
			       0xca, 0x04, 0x7a, 0xb7 };

      XChangeProperty(display, res->win, WM_HINTS, WM_HINTS, 32,
		      PropModeReplace, test,5);
    }
  }


  // skip taskbar/pager, sticky, below

  {
    Atom atoms[4];
    Atom WM_HINTS;

    Atom atom = XInternAtom(display,"ATOM",1);

    atoms[0] = XInternAtom(display, "_NET_WM_STATE_BELOW",0);
    atoms[1] = XInternAtom(display, "_NET_WM_STATE_STICKY",0);
    atoms[2] = XInternAtom(display, "_NET_WM_STATE_SKIP_PAGER",0);
    atoms[3] = XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR",0);


    WM_HINTS = XInternAtom(display, "_NET_WM_STATE", True);

    XChangeProperty(display, res->win, WM_HINTS, atom, 32,
		    PropModeReplace,
    		    (unsigned char*) &atoms, 4);
  }

  // on all workspace


  {
    Atom atom = XInternAtom(display,"CARDINAL",0);
    Atom wm   = XInternAtom(display, "_NET_WM_DESKTOP", True);
    Atom val = 0xFFFFFFFF;

    XChangeProperty(display, res->win, wm, atom, 32,
		    PropModeReplace,
		    (unsigned char*) &val, 1);
  }

  
  
  XStoreName(display, res->win, name);
  XMapWindow(display, res->win);
  XFlush(display);

  res->buff = XCreatePixmap(display, res->win, res->win_size_x, res->win_size_y,
  		       ScreenOfDisplay(display, DefaultScreen (display))->root_depth);


  res->xft_draw = XftDrawCreate(display, res->buff,
		       DefaultVisual(display, DefaultScreen (display)),
		       DefaultColormap(display, DefaultScreen (display)));

  return res;
}


void gosd_init() {
  char *display_name = getenv("DISPLAY");
  display = XOpenDisplay(display_name);
}

Pixmap gosd_get_buffer(const win_t *w) {
  return w->buff;
}

Window gosd_get_window(const win_t *w) {
  return w->win;
}

Display *gosd_get_display() {
  return display;
}

